#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import decimal as _builtin_decimal
import json as _json
import warnings
from collections import OrderedDict
from datetime import datetime as _datetime, timedelta as _timedelta, date as _date

from . import compat, utils
from .compat import six, DECIMAL_TYPES, decimal as _decimal, east_asian_len, Monthdelta
from .config import options
from .lib.xnamedtuple import xnamedtuple

try:
    from pandas import NA as _pd_na
    pd_na_type = type(_pd_na)
except ImportError:
    pd_na_type = None

force_py = options.force_py
force_c = options.force_c


class Column(object):
    def __init__(
        self, name=None, typo=None, comment=None, label=None, nullable=True, **kw
    ):
        self.name = utils.to_str(name)
        self.type = validate_data_type(
            typo if typo is not None else kw.pop("type", None)
        )
        self.comment = comment
        self.label = label
        self.nullable = nullable
        if kw:
            raise TypeError("Arguments not supported for Column: %s" % list(kw))

    def __repr__(self):
        not_null_str = ", not null" if not self.nullable else ""
        return '<column {0}, type {1}{2}>'.format(
            utils.to_str(self.name), self.type.name.lower(), not_null_str
        )

    def __hash__(self):
        return hash((type(self), self.name, self.type, self.comment, self.label))


class Partition(Column):
    def __repr__(self):
        return '<partition {0}, type {1}>'.format(utils.to_str(self.name), self.type.name.lower())


class _CallableList(list):
    """Make sure keys and values properties also callable"""
    def __call__(self):
        return self


class PartitionSpec(object):
    def __init__(self, spec=None):
        self.kv = OrderedDict()

        if isinstance(spec, PartitionSpec):
            self.kv = spec.kv.copy()
        elif isinstance(spec, dict):
            self.kv = OrderedDict(spec)
        elif isinstance(spec, six.string_types):
            splits = spec.split(',')
            for sp in splits:
                kv = sp.split('=')
                if len(kv) != 2:
                    raise ValueError(
                        'Invalid partition spec: a partition spec should '
                        'look like "part1=v1,part2=v2"'
                    )

                k, v = kv[0].strip(), kv[1].strip().strip('\'"')

                if len(k) == 0 or len(v) == 0:
                    raise ValueError('Invalid partition spec')
                if k in self.kv:
                    raise ValueError(
                        'Invalid partition spec: found duplicate partition key ' + k
                    )

                self.kv[k] = v
        elif spec is not None:
            raise TypeError("Cannot accept spec %r" % spec)

    def __setitem__(self, key, value):
        self.kv[key] = value

    def __getitem__(self, key):
        return self.kv[key]

    def __len__(self):
        return len(self.kv)

    @property
    def is_empty(self):
        return len(self) == 0

    @property
    def keys(self):
        return _CallableList(self.kv.keys())

    @property
    def values(self):
        return _CallableList(self.kv.values())

    def items(self):
        for k, v in self.kv.items():
            yield k, v

    def __contains__(self, key):
        return key in self.kv

    def __str__(self):
        return ','.join("%s='%s'" % (k, v) for k, v in six.iteritems(self.kv))

    def __repr__(self):
        return '<PartitionSpec %s>' % str(self)

    def __hash__(self):
        return hash(str(self))

    def __eq__(self, other):
        if not isinstance(other, PartitionSpec):
            other = PartitionSpec(other)

        return str(self) == str(other)


class Schema(object):
    def __init__(self, names, types):
        self._init(names, types)

    def _init(self, names, types):
        if not isinstance(names, list):
            names = list(names)
        self.names = names
        self.types = [validate_data_type(t) for t in types]

        self._name_indexes = {n: i for i, n in enumerate(self.names)}

        if len(self._name_indexes) < len(self.names):
            duplicates = [n for n in self._name_indexes if self.names.count(n) > 1]
            raise ValueError('Duplicate column names: %s' % ', '.join(duplicates))

        self._snapshot = None

    def __repr__(self):
        return self._repr()

    def __len__(self):
        return len(self.names)

    def __contains__(self, name):
        return utils.to_str(name) in self._name_indexes

    def _repr(self):
        buf = six.StringIO()
        names = [self._to_printable(n) for n in self.names]
        space = 2 * max(len(it) for it in names)
        for name, tp in zip(names, self.types):
            buf.write('\n{0}{1}'.format(name.ljust(space), repr(tp)))

        return 'Schema {{{0}\n}}'.format(utils.indent(buf.getvalue(), 2))

    def __hash__(self):
        return hash((type(self), tuple(self.names), tuple(self.types)))

    def __eq__(self, other):
        if not isinstance(other, Schema):
            return False
        return self.names == other.names and self.types == self.types

    def get_type(self, name):
        return self.types[self._name_indexes[utils.to_str(name)]]

    def append(self, name, typo):
        names = self.names + [name, ]
        types = self.types + [validate_data_type(typo), ]
        return Schema(names, types)

    def extend(self, schema):
        names = self.names + schema.names
        types = self.types + schema.types
        return Schema(names, types)


class OdpsSchema(Schema):
    def __init__(self, columns=None, partitions=None):
        self._columns = columns
        self._partitions = partitions

        if self._columns:
            super(OdpsSchema, self).__init__(
                *compat.lzip(*[(c.name, c.type) for c in self._columns]))
        else:
            super(OdpsSchema, self).__init__([], [])

        if self._partitions:
            self._partition_schema = Schema(
                *compat.lzip(*[(c.name, c.type) for c in self._partitions]))
        else:
            self._partition_schema = Schema([], [])

    def __len__(self):
        return super(OdpsSchema, self).__len__() + len(self._partition_schema)

    def __setattr__(self, key, value):
        if key == '_columns' and value and not getattr(self, 'names', None) and \
                not getattr(self, 'types', None):
            names = [c.name for c in value]
            types = [c.type for c in value]
            self._init(names, types)
        elif key == '_partitions' and value:
            self._partition_schema = Schema(
                *compat.lzip(*[(c.name, c.type) for c in value]))
        object.__setattr__(self, key, value)

    def __contains__(self, name):
        return super(OdpsSchema, self).__contains__(name) or \
               utils.to_str(name) in self._partition_schema

    def __eq__(self, other):
        if not isinstance(other, OdpsSchema):
            return False

        return super(OdpsSchema, self).__eq__(other) and \
            self._partition_schema == other._partition_schema

    def __hash__(self):
        return hash((type(self), tuple(self.names), tuple(self.types),
                     self._partition_schema))

    def __getitem__(self, item):
        if isinstance(item, six.integer_types):
            n_columns = len(self._name_indexes)
            if item < n_columns:
                return self._columns[item]
            elif item < len(self):
                return self._partitions[item-n_columns]
            else:
                raise IndexError('Index out of range')
        elif isinstance(item, six.string_types):
            item = utils.to_str(item)
            if item in self._name_indexes:
                idx = self._name_indexes[item]
                return self[idx]
            elif item in self._partition_schema:
                idx = self._partition_schema._name_indexes[item]
                n_columns = len(self._name_indexes)
                return self[n_columns+idx]
            else:
                raise ValueError('Unknown column name: %s' % item)
        elif isinstance(item, (list, tuple)):
            return [self[it] for it in item]
        else:
            return self.columns[item]

    def _repr(self):
        buf = six.StringIO()

        name_dict = dict([(col.name, utils.str_to_printable(col.name)) for col in self.columns])
        name_display_lens = dict([(k, east_asian_len(utils.to_text(v), encoding=options.display.encoding))
                                  for k, v in six.iteritems(name_dict)])
        name_space = 2 * max(six.itervalues(name_display_lens))
        type_space = 2 * max(len(repr(col.type)) for col in self.columns)
        has_not_null = any(not col.nullable for col in self.columns)

        not_empty = lambda field: field is not None and len(field.strip()) > 0

        buf.write('odps.Schema {\n')
        cols_strs = []
        for col in self._columns:
            pad_spaces = name_space - name_display_lens[col.name]
            not_null = "not null" if not col.nullable else " " * 8
            cols_strs.append('{0}{1}{2}{3}'.format(
                utils.to_str(name_dict[col.name] + ' ' * pad_spaces),
                repr(col.type).ljust(type_space),
                not_null + " " * 4 if has_not_null else "",
                '# {0}'.format(utils.to_str(col.comment)) if not_empty(col.comment) else ''
            ))
        buf.write(utils.indent('\n'.join(cols_strs), 2))
        buf.write('\n')
        buf.write('}\n')

        if self._partitions:
            buf.write('Partitions {\n')

            partition_strs = []
            for partition in self._partitions:
                partition_strs.append('{0}{1}{2}'.format(
                    utils.to_str(name_dict[partition.name].ljust(name_space)),
                    repr(partition.type).ljust(type_space),
                    '# {0}'.format(utils.to_str(partition.comment)) if not_empty(partition.comment) else ''
                ))
            buf.write(utils.indent('\n'.join(partition_strs), 2))
            buf.write('\n')
            buf.write('}\n')

        return buf.getvalue()

    def build_snapshot(self):
        if not options.force_py:
            if not self._columns:
                return None

            try:
                from .src.types_c import SchemaSnapshot
                self._snapshot = SchemaSnapshot(self)
            except ImportError:
                pass
        return self._snapshot

    @property
    def simple_columns(self):
        return self._columns

    @property
    def columns(self):
        partitions = self._partitions or []
        return self._columns + partitions

    @property
    def partitions(self):
        try:
            return self._partitions
        except AttributeError:
            return []

    @utils.deprecated('use simple_columns property instead')
    def get_columns(self):
        return self._columns

    @utils.deprecated('use partitions property instead')
    def get_partitions(self):
        return self._partitions

    def get_column(self, name):
        index = self._name_indexes.get(utils.to_str(name))
        if index is None:
            raise ValueError('Column %s does not exists' % name)
        return self._columns[index]

    def get_partition(self, name):
        index = self._partition_schema._name_indexes.get(utils.to_str(name))
        if index is None:
            raise ValueError('Partition %s does not exists' % name)
        return self._partitions[index]

    def is_partition(self, name):
        try:
            name = name.name
        except AttributeError:
            pass
        return name in self._partition_schema._name_indexes

    def get_type(self, name):
        if name in self._name_indexes:
            return super(OdpsSchema, self).get_type(name)
        elif name in self._partition_schema:
            return self._partition_schema.get_type(name)
        raise ValueError('Column does not exist: %s' % name)

    def update(self, columns, partitions):
        self._columns = columns
        self._partitions = partitions

        names = map(lambda c: c.name, self._columns)
        types = map(lambda c: c.type, self._columns)

        self._init(names, types)
        if self._partitions:
            self._partition_schema = Schema(
                *compat.lzip(*[(c.name, c.type) for c in self._partitions]))
        else:
            self._partition_schema = Schema([], [])

    def extend(self, schema):
        if isinstance(schema, Schema):
            ext_cols = [Column(n, tp) for n, tp in zip(schema.names, schema.types)]
            ext_parts = []
        else:
            ext_cols = schema.simple_columns
            ext_parts = schema.partitions
        return type(self)(
            columns=self.simple_columns + ext_cols,
            partitions=self.partitions + ext_parts,
        )

    def to_ignorecase_schema(self):
        cols = [Column(col.name.lower(), col.type, col.comment, col.label)
                for col in self._columns]
        parts = None
        if self._partitions:
            parts = [Partition(part.name.lower(), part.type, part.comment, part.label)
                     for part in self._partitions]

        return type(self)(columns=cols, partitions=parts)

    @classmethod
    def from_lists(cls, names, types, partition_names=None, partition_types=None):
        columns = [Column(name=name, typo=typo) for name, typo in zip(names, types)]
        if partition_names is not None and partition_types is not None:
            partitions = [Partition(name=name, typo=typo)
                          for name, typo in zip(partition_names, partition_types)]
        else:
            partitions = None
        return cls(columns=columns, partitions=partitions)

    @classmethod
    def from_dict(cls, fields_dict, partitions_dict=None):
        fields = compat.lkeys(fields_dict)
        fields_types = compat.lvalues(fields_dict)
        partitions = compat.lkeys(partitions_dict) \
            if partitions_dict is not None else None
        partitions_types = compat.lvalues(partitions_dict) \
            if partitions_dict is not None else None

        return cls.from_lists(fields, fields_types,
                              partition_names=partitions,
                              partition_types=partitions_types)

    def get_table_ddl(self, table_name='table_name', with_comments=True):
        from .models.table import Table
        return Table.gen_create_table_sql(table_name, self, with_column_comments=with_comments)


class RecordMeta(type):
    record_types = set()

    def __new__(mcs, name, bases, dct):
        inst = super(RecordMeta, mcs).__new__(mcs, name, bases, dct)
        mcs.record_types.add(inst)
        return inst

    def __instancecheck__(cls, instance):
        return isinstance(instance, RecordReprMixin)


def is_record(obj):
    return type(obj) in RecordMeta.record_types


class BaseRecord(object):

    # set __slots__ to save memory in the situation that records' size may be quite large
    __slots__ = '_values', '_columns', '_name_indexes', '_max_field_size'

    def __init__(self, columns=None, schema=None, values=None, max_field_size=None):
        if columns is not None:
            self._columns = columns
            self._name_indexes = {col.name: i for i, col in enumerate(self._columns)}
        else:
            self._columns = schema.columns
            self._name_indexes = schema._name_indexes

        self._max_field_size = max_field_size

        if self._columns is None:
            raise ValueError('Either columns or schema should not be provided')

        self._values = [None, ] * len(self._columns)
        if values is not None:
            self._sets(values)

    def _mode(self):
        return 'py'

    def _exclude_partition_columns(self):
        return [col for col in self._columns if not isinstance(col, Partition)]

    def _get(self, i):
        return self._values[i]

    def _set(self, i, value):
        data_type = self._columns[i].type
        val = validate_value(value, data_type, max_field_size=self._max_field_size)
        self._values[i] = val

    get = _get  # to keep compatible
    set = _set  # to keep compatible

    def _sets(self, values):
        if len(values) != len(self._columns) and \
                        len(values) != len(self._exclude_partition_columns()):
            raise ValueError('The values set to records are against the schema, '
                             'expect len %s, got len %s' % (len(self._columns), len(values)))
        [self._set(i, value) for i, value in enumerate(values)]

    def __getitem__(self, item):
        if isinstance(item, six.string_types):
            return self.get_by_name(item)
        elif isinstance(item, (list, tuple)):
            return [self[it] for it in item]
        return self._values[item]

    def __setitem__(self, key, value):
        if isinstance(key, six.string_types):
            self.set_by_name(key, value)
        else:
            self._set(key, value)

    def __getattr__(self, item):
        if item == '_name_indexes':
            return object.__getattribute__(self, item)
        if hasattr(self, '_name_indexes') and item in self._name_indexes:
            return self.get_by_name(item)
        return object.__getattribute__(self, item)

    def __setattr__(self, key, value):
        if hasattr(self, '_name_indexes') and key in self._name_indexes:
            self.set_by_name(key, value)
        else:
            object.__setattr__(self, key, value)

    def get_by_name(self, name):
        i = self._name_indexes[name]
        return self._values[i]

    def set_by_name(self, name, value):
        i = self._name_indexes[name]
        self._set(i, value)

    def __len__(self):
        return len(self._columns)

    def __contains__(self, item):
        return item in self._name_indexes

    def __iter__(self):
        for i, col in enumerate(self._columns):
            yield (col.name, self[i])

    @property
    def values(self):
        return self._values

    @property
    def n_columns(self):
        return len(self._columns)

    def get_columns_count(self):  # compatible
        return self.n_columns


class RecordReprMixin(object):
    def __repr__(self):
        buf = six.StringIO()

        buf.write('odps.Record {\n')

        space = 2 * max(len(it.name) for it in self._columns)
        content = '\n'.join(
            ['{0}{1}'.format(col.name.ljust(space), repr(value))
             for col, value in zip(self._columns, self._values)])
        buf.write(utils.indent(content, 2))

        buf.write('\n}')

        return buf.getvalue()

    def __hash__(self):
        return hash((type(self), tuple(self._columns), tuple(self._values)))

    def __eq__(self, other):
        if not is_record(other):
            return False

        return self._columns == other._columns and self._values == other._values


class Record(six.with_metaclass(RecordMeta, RecordReprMixin, BaseRecord)):
    """
    A record generally means the data of a single line in a table.

    :Example:

    >>> schema = TableSchema.from_lists(['name', 'id'], ['string', 'string'])
    >>> record = Record(schema=schema, values=['test', 'test2'])
    >>> record[0] = 'test'
    >>> record[0]
    >>> 'test'
    >>> record['name']
    >>> 'test'
    >>> record[0:2]
    >>> ('test', 'test2')
    >>> record[0, 1]
    >>> ('test', 'test2')
    >>> record['name', 'id']
    >>> for field in record:
    >>>     print(field)
    ('name', u'test')
    ('id', u'test2')
    >>> len(record)
    2
    >>> 'name' in record
    True
    """


class DataType(object):
    """
    Abstract data type
    """
    _singleton = True
    _type_id = -1
    __slots__ = 'nullable',

    def __new__(cls, *args, **kwargs):
        if cls._singleton:
            if not hasattr(cls, '_instance'):
                cls._instance = object.__new__(cls)
                cls._hash = hash(cls)
            return cls._instance
        else:
            return object.__new__(cls)

    def __init__(self, nullable=True):
        self.nullable = nullable

    def __call__(self, nullable=True):
        return self._factory(nullable=nullable)

    def _factory(self, nullable=True):
        return type(self)(nullable=nullable)

    def __ne__(self, other):
        return not (self == other)

    def __eq__(self, other):
        try:
            return self._equals(other)
        except (TypeError, ValueError):
            return False

    def _equals(self, other):
        if self is other:
            return True

        other = validate_data_type(other)

        if self.nullable != other.nullable:
            return False
        if type(self) == type(other):
            return True
        return isinstance(other, type(self))

    def __hash__(self):
        return self._hash

    @property
    def name(self):
        return type(self).__name__.lower()

    def __repr__(self):
        if self.nullable:
            return self.name
        return '{0}[non-nullable]'.format(self.name)

    def __str__(self):
        return self.name.upper()

    def can_implicit_cast(self, other):
        if isinstance(other, six.string_types):
            other = validate_data_type(other)

        return isinstance(self, type(other))

    def can_explicit_cast(self, other):
        return self.can_implicit_cast(other)

    def validate_value(self, val, max_field_size=None):
        # directly return True means without checking
        return True

    def _can_cast_or_throw(self, value, data_type):
        if not self.can_implicit_cast(data_type):
            raise ValueError('Cannot cast value(%s) from type(%s) to type(%s)' % (
                value, data_type, self))

    def cast_value(self, value, data_type):
        raise NotImplementedError


class OdpsPrimitive(DataType):
    __slots__ = ()


class BaseInteger(OdpsPrimitive):
    __slots__ = ()

    _type_id = 0
    _bounds = None
    _store_bytes = None

    def can_implicit_cast(self, other):
        if isinstance(other, six.string_types):
            other = validate_data_type(other)

        if isinstance(other, (BaseFloat, String, Decimal)):
            return True
        if isinstance(other, BaseInteger):
            return self._store_bytes >= other._store_bytes
        return super(BaseInteger, self).can_implicit_cast(other)

    def validate_value(self, val, max_field_size=None):
        if val is None and self.nullable:
            return True
        smallest, largest = self._bounds
        if smallest <= val <= largest:
            return True
        raise ValueError('InvalidData: Bigint(%s) out of range' % val)

    def cast_value(self, value, data_type):
        self._can_cast_or_throw(value, data_type)

        return int(value)


class Tinyint(BaseInteger):
    _bounds = (-128, 127)
    _store_bytes = 1


class Smallint(BaseInteger):
    _bounds = (-32768, 32767)
    _store_bytes = 2


class Int(BaseInteger):
    _bounds = (-2147483648, 2147483647)
    _store_bytes = 4


class Bigint(BaseInteger):
    _bounds = (-9223372036854775808, 9223372036854775807)
    _store_bytes = 8


class BaseFloat(OdpsPrimitive):
    __slots__ = ()
    _store_bytes = None

    def can_implicit_cast(self, other):
        if isinstance(other, six.string_types):
            other = validate_data_type(other)

        if isinstance(other, (BaseInteger, String, Decimal)):
            return True
        if isinstance(other, BaseFloat):
            return self._store_bytes >= other._store_bytes
        return super(BaseFloat, self).can_implicit_cast(other)

    def cast_value(self, value, data_type):
        self._can_cast_or_throw(value, data_type)

        return float(value)


class Float(BaseFloat):
    _store_bytes = 4
    _type_id = 6


class Double(BaseFloat):
    _store_bytes = 8
    _type_id = 1


class String(OdpsPrimitive):
    __slots__ = ()

    _type_id = 2
    _max_length = 8 * 1024 * 1024  # 8M

    def can_implicit_cast(self, other):
        if isinstance(other, six.string_types):
            other = validate_data_type(other)

        if isinstance(other, (BaseInteger, BaseFloat, Datetime, Decimal, Binary, Json)):
            return True
        return super(String, self).can_implicit_cast(other)

    def validate_value(self, val, max_field_size=None):
        if val is None and self.nullable:
            return True
        max_field_size = max_field_size or self._max_length
        if len(val) <= max_field_size:
            return True
        raise ValueError(
            "InvalidData: Length of string(%s) is more than %sM.'" %
            (val, max_field_size / (1024 ** 2))
        )

    def cast_value(self, value, data_type):
        self._can_cast_or_throw(value, data_type)

        if isinstance(data_type, Datetime):
            return value.strftime('%Y-%m-%d %H:%M:%S')
        if options.tunnel.string_as_binary:
            val = utils.to_binary(value)
        else:
            val = utils.to_text(value)
        return val


class Datetime(OdpsPrimitive):
    __slots__ = ()
    _type_id = 3

    def can_implicit_cast(self, other):
        if isinstance(other, six.string_types):
            other = validate_data_type(other)

        if isinstance(other, (BaseTimestamp, Datetime, String)):
            return True
        return super(Datetime, self).can_implicit_cast(other)

    def cast_value(self, value, data_type):
        self._can_cast_or_throw(value, data_type)

        if isinstance(data_type, String):
            return _datetime.strptime(value, '%Y-%m-%d %H:%M:%S')
        elif isinstance(data_type, BaseTimestamp):
            return value.to_pydatetime()
        return value


class Date(OdpsPrimitive):
    __slots__ = ()
    _type_id = 11

    def can_implicit_cast(self, other):
        if isinstance(other, six.string_types):
            other = validate_data_type(other)

        if isinstance(other, (BaseTimestamp, Datetime, String)):
            return True
        return super(Date, self).can_implicit_cast(other)

    def cast_value(self, value, data_type):
        self._can_cast_or_throw(value, data_type)

        if isinstance(data_type, String):
            datetime = _datetime.strptime(value, '%Y-%m-%d')
            return _date(datetime.year, datetime.month, datetime.day)
        elif isinstance(data_type, BaseTimestamp):
            return value.to_pydatetime().date()
        return value


class Boolean(OdpsPrimitive):
    __slots__ = ()
    _type_id = 4

    def cast_value(self, value, data_type):
        self._can_cast_or_throw(value, data_type)
        return value


class Binary(OdpsPrimitive):
    __slots__ = ()
    _type_id = 7
    _max_length = 8 * 1024 * 1024  # 8M

    def can_implicit_cast(self, other):
        if isinstance(other, six.string_types):
            other = validate_data_type(other)

        if isinstance(other, (BaseInteger, BaseFloat, Datetime, Decimal, String)):
            return True
        return super(Binary, self).can_implicit_cast(other)

    def validate_value(self, val, max_field_size=None):
        if val is None and self.nullable:
            return True
        max_field_size = max_field_size or self._max_length
        if len(val) <= max_field_size:
            return True
        raise ValueError(
            "InvalidData: Length of binary(%s) is more than %sM.'" %
            (val, max_field_size / (1024 ** 2))
        )

    def cast_value(self, value, data_type):
        self._can_cast_or_throw(value, data_type)

        if isinstance(data_type, Datetime):
            return value.strftime('%Y-%m-%d %H:%M:%S')
        return utils.to_binary(value)


class BaseTimestamp(OdpsPrimitive):
    __slots__ = ()
    _type_id = 8

    def can_implicit_cast(self, other):
        if isinstance(other, six.string_types):
            other = validate_data_type(other)

        if isinstance(other, (BaseTimestamp, Datetime, String)):
            return True
        return super(BaseTimestamp, self).can_implicit_cast(other)

    def cast_value(self, value, data_type):
        self._can_cast_or_throw(value, data_type)
        try:
            import pandas as pd
        except ImportError:
            raise ImportError('To use TIMESTAMP in pyodps, you need to install pandas.')

        if isinstance(data_type, String):
            return pd.to_datetime(value)
        elif isinstance(data_type, Datetime):
            return pd.Timestamp(value)
        return value


class Timestamp(BaseTimestamp):
    _type_id = 8


class TimestampNTZ(BaseTimestamp):
    _type_id = 13

    @property
    def name(self):
        return "timestamp_ntz"


class IntervalDayTime(OdpsPrimitive):
    __slots__ = ()
    _type_id = 9

    def can_implicit_cast(self, other):
        if isinstance(other, six.string_types):
            other = validate_data_type(other)

        if isinstance(other, (BaseTimestamp, Datetime, String)):
            return True
        return super(IntervalDayTime, self).can_implicit_cast(other)

    @property
    def name(self):
        return 'interval_day_time'

    def cast_value(self, value, data_type):
        self._can_cast_or_throw(value, data_type)
        try:
            import pandas as pd
        except ImportError:
            raise ImportError('To use INTERVAL_DAY_TIME in pyodps, you need to install pandas.')

        if isinstance(value, float):
            return pd.Timedelta(seconds=value)
        elif isinstance(value, _timedelta):
            return pd.Timedelta(value)
        return value


class IntervalYearMonth(OdpsPrimitive):
    __slots__ = ()
    _type_id = 10

    def can_implicit_cast(self, other):
        if isinstance(other, six.string_types):
            other = validate_data_type(other)

        if isinstance(other, (String, BaseInteger)):
            return True
        return super(IntervalYearMonth, self).can_implicit_cast(other)

    @property
    def name(self):
        return 'interval_year_month'

    def cast_value(self, value, data_type):
        self._can_cast_or_throw(value, data_type)

        if isinstance(value, (int, compat.long_type, six.string_types)):
            return Monthdelta(value)
        return value


class CompositeMixin(DataType):
    _singleton = False

    @classmethod
    def parse_composite(cls, args):
        raise NotImplementedError

    def validate_composite_values(self, value):
        raise NotImplementedError


class SizeLimitedString(String, CompositeMixin):
    _singleton = False
    __slots__ = 'nullable', 'size_limit', '_hash'
    _max_length = 65535

    def __init__(self, size_limit, nullable=True):
        super(SizeLimitedString, self).__init__(nullable=nullable)
        if size_limit > self._max_length:
            raise ValueError("InvalidData: Length of varchar(%d) is larger than %d." %
                             (size_limit, self._max_length))
        self.size_limit = size_limit

    @property
    def name(self):
        return '{0}({1})'.format(type(self).__name__.lower(),
                                 self.size_limit)

    def _equals(self, other):
        if isinstance(other, six.string_types):
            other = validate_data_type(other)

        return DataType._equals(self, other) and \
               self.size_limit == other.size_limit

    def __hash__(self):
        if not hasattr(self, '_hash'):
            self._hash = hash((type(self), self.nullable, self.size_limit))
        return self._hash

    def validate_value(self, val, max_field_size=None):
        if val is None and self.nullable:
            return True
        if len(val) <= self.size_limit:
            return True
        raise ValueError(
            "InvalidData: Length of string(%d) is more than %sM.'" %
            (len(val), self.size_limit))

    @classmethod
    def parse_composite(cls, args):
        if len(args) != 1:
            raise ValueError('%s() only accept one length argument.' % cls.__name__.upper())
        try:
            return cls(int(args[0]))
        except TypeError:
            raise ValueError('%s() only accept an integer length argument.' % cls.__name__.upper())

    def validate_composite_values(self, value):
        self.validate_value(value)
        return self.cast_value(value, self)


class Varchar(SizeLimitedString):
    def can_implicit_cast(self, other):
        if isinstance(other, six.string_types):
            other = validate_data_type(other)

        if isinstance(other, (BaseInteger, BaseFloat, Datetime, Decimal, String, Binary)):
            return True
        return isinstance(other, (Char, Varchar)) and \
               self.size_limit >= other.size_limit and \
               self.nullable == other.nullable


class Char(SizeLimitedString):
    def can_implicit_cast(self, other):
        if isinstance(other, six.string_types):
            other = validate_data_type(other)

        if isinstance(other, (BaseInteger, BaseFloat, Datetime, Decimal, String, Binary)):
            return True
        return isinstance(other, (Char, Varchar)) and \
               self.size_limit >= other.size_limit and \
               self.nullable == other.nullable


class Decimal(CompositeMixin):
    __slots__ = 'nullable', 'precision', 'scale', '_hash'
    _type_id = 5

    _has_other_decimal_type = len(DECIMAL_TYPES) > 1

    _max_precision = 54
    _max_scale = 18
    _decimal_ctx = _decimal.Context(prec=_max_precision)

    def __init__(self, precision=None, scale=None, nullable=True):
        super(Decimal, self).__init__(nullable=nullable)
        if precision and precision > self._max_precision:
            raise ValueError(
                "InvalidData: Precision(%d) is larger than %d."
                % (precision, self._max_precision)
            )
        if scale and scale > self._max_scale:
            raise ValueError(
                "InvalidData: Scale(%d) is larger than %d." % (scale, self._max_scale)
            )
        if precision is None and scale is not None:
            raise ValueError('InvalidData: Scale should be provided along with precision.')
        self.precision = precision
        self.scale = scale
        self._scale_decimal = _decimal.Decimal(
            "1e%d" % -(scale if scale is not None else self._max_scale)
        )

    @property
    def name(self):
        if self.precision is None:
            return type(self).__name__.lower()
        elif self.scale is None:
            return '{0}({1})'.format(type(self).__name__.lower(),
                                     self.precision)
        else:
            return '{0}({1},{2})'.format(type(self).__name__.lower(),
                                         self.precision, self.scale)

    def __hash__(self):
        if not hasattr(self, '_hash'):
            self._hash = hash((type(self), self.nullable, self.precision, self.scale))
        return self._hash

    def _equals(self, other):
        if isinstance(other, six.string_types):
            other = validate_data_type(other)

        return (
            DataType._equals(self, other)
            and self.precision == other.precision
            and self.scale == other.scale
        )

    def can_implicit_cast(self, other):
        if isinstance(other, six.string_types):
            other = validate_data_type(other)

        if isinstance(other, (BaseInteger, BaseFloat, String)):
            return True
        return super(Decimal, self).can_implicit_cast(other)

    def validate_value(self, val, max_field_size=None):
        if val is None and self.nullable:
            return True

        if (
            self._has_other_decimal_type
            and not isinstance(val, _decimal.Decimal)
            and isinstance(val, DECIMAL_TYPES)
        ):
            val = _decimal.Decimal(str(val))

        precision = self.precision if self.precision is not None else self._max_precision
        scale = self.scale if self.scale is not None else self._max_scale
        scaled_val = val.quantize(
            self._scale_decimal, _decimal.ROUND_HALF_UP, self._decimal_ctx
        )
        int_len = len(str(scaled_val)) - scale - 1
        if int_len > precision:
            raise ValueError(
                'decimal value %s overflow, max integer digit number is %s.' %
                (val, precision)
            )
        return True

    def cast_value(self, value, data_type):
        self._can_cast_or_throw(value, data_type)

        if six.PY3 and isinstance(value, six.binary_type):
            value = value.decode('utf-8')
        return _builtin_decimal.Decimal(value)

    @classmethod
    def parse_composite(cls, args):
        if len(args) > 2:
            raise ValueError('%s() accepts no more than two arguments.' % cls.__name__.upper())
        try:
            return cls(*[int(v) for v in args])
        except TypeError:
            raise ValueError('%s() only accept integers as arguments.' % cls.__name__.upper())

    def validate_composite_values(self, value):
        if value is None and self.nullable:
            return value
        if not isinstance(value, _builtin_decimal.Decimal):
            value = self.cast_value(value, infer_primitive_data_type(value))
        self.validate_value(value)
        return value


class Array(CompositeMixin):
    __slots__ = 'nullable', 'value_type', '_hash'

    def __init__(self, value_type, nullable=True):
        super(Array, self).__init__(nullable=nullable)
        value_type = validate_data_type(value_type)
        self.value_type = value_type

    @property
    def name(self):
        return '{0}<{1}>'.format(type(self).__name__.lower(),
                                 self.value_type.name)

    def _equals(self, other):
        if isinstance(other, six.string_types):
            other = validate_data_type(other)

        return DataType._equals(self, other) and \
            self.value_type == other.value_type

    def __hash__(self):
        if not hasattr(self, '_hash'):
            self._hash = hash((type(self), self.nullable, hash(self.value_type)))
        return self._hash

    def can_implicit_cast(self, other):
        if isinstance(other, six.string_types):
            other = validate_data_type(other)

        return isinstance(other, Array) and \
            self.value_type == other.value_type and \
            self.nullable == other.nullable

    def cast_value(self, value, data_type):
        self._can_cast_or_throw(value, data_type)
        return value

    @classmethod
    def parse_composite(cls, args):
        if len(args) != 1:
            raise ValueError('%s<> should be supplied with exactly one type.' % cls.__name__.upper())
        return cls(args[0])

    def validate_composite_values(self, value):
        if value is None and self.nullable:
            return value
        if not isinstance(value, list):
            raise ValueError('Array data type requires `list`, instead of %s' % value)
        element_data_type = self.value_type
        return [validate_value(element, element_data_type) for element in value]


class Map(CompositeMixin):
    __slots__ = 'nullable', 'key_type', 'value_type', '_hash'

    def __init__(self, key_type, value_type, nullable=True):
        super(Map, self).__init__(nullable=nullable)
        key_type = validate_data_type(key_type)
        value_type = validate_data_type(value_type)
        self.key_type = key_type
        self.value_type = value_type

    @property
    def name(self):
        return '{0}<{1},{2}>'.format(type(self).__name__.lower(),
                                     self.key_type.name,
                                     self.value_type.name)

    def _equals(self, other):
        if isinstance(other, six.string_types):
            other = validate_data_type(other)

        return DataType._equals(self, other) and \
            self.key_type == other.key_type and \
            self.value_type == other.value_type

    def __hash__(self):
        if not hasattr(self, '_hash'):
            self._hash = hash((type(self), self.nullable,
                               hash(self.key_type), hash(self.value_type)))
        return self._hash

    def can_implicit_cast(self, other):
        if isinstance(other, six.string_types):
            other = validate_data_type(other)

        return isinstance(other, Map) and \
            self.key_type == other.key_type and \
            self.value_type == other.value_type and \
            self.nullable == other.nullable

    def cast_value(self, value, data_type):
        self._can_cast_or_throw(value, data_type)
        return value

    @classmethod
    def parse_composite(cls, args):
        if len(args) != 2:
            raise ValueError('%s<> should be supplied with exactly two types.' % cls.__name__.upper())
        return cls(*args)

    def validate_composite_values(self, value):
        if value is None and self.nullable:
            return value
        if not isinstance(value, dict):
            raise ValueError('Map data type requires `dict`, instead of %s' % value)
        key_data_type = self.key_type
        value_data_type = self.value_type

        convert = lambda k, v: (validate_value(k, key_data_type),
                                validate_value(v, value_data_type))
        return OrderedDict(convert(k, v) for k, v in six.iteritems(value))


class Struct(CompositeMixin):
    __slots__ = 'nullable', 'field_types', '_hash'

    def __init__(self, field_types, nullable=True):
        super(Struct, self).__init__(nullable=nullable)
        self.field_types = OrderedDict()
        if isinstance(field_types, dict):
            field_types = six.iteritems(field_types)
        for k, v in field_types:
            self.field_types[k] = validate_data_type(v)
        self.namedtuple_type = xnamedtuple(
            "StructNamedTuple", list(self.field_types.keys())
        )

        self._struct_as_dict = options.struct_as_dict
        if self._struct_as_dict:
            warnings.warn(
                "Representing struct values as dicts is now deprecated. Try config "
                "`options.struct_as_dict=False` and return structs as named tuples "
                "instead.", DeprecationWarning
            )
            utils.add_survey_call("options.struct_as_dict")

    @property
    def name(self):
        parts = ','.join('`%s`:%s' % (k, v.name) for k, v in six.iteritems(self.field_types))
        return '{0}<{1}>'.format(type(self).__name__.lower(), parts)

    def _equals(self, other):
        if isinstance(other, six.string_types):
            other = validate_data_type(other)

        return isinstance(other, Struct) and \
            len(self.field_types) == len(other.field_types) and \
            all(self.field_types[k] == other.field_types[k] for k in six.iterkeys(self.field_types))

    def __hash__(self):
        if not hasattr(self, '_hash'):
            fields_hash = hash(tuple((hash(k), hash(v)) for k, v in six.iteritems(self.field_types)))
            self._hash = hash((type(self), self.nullable, fields_hash))
        return self._hash

    def can_implicit_cast(self, other):
        if isinstance(other, six.string_types):
            other = validate_data_type(other)

        return isinstance(other, Struct) and self == other and \
               self.nullable == other.nullable

    def cast_value(self, value, data_type):
        self._can_cast_or_throw(value, data_type)
        return value

    @classmethod
    def parse_composite(cls, args):
        if any(not isinstance(a, tuple) and ':' not in a
               for a in args):
            raise ValueError('Every field defined in STRUCT should be given a name.')

        def conv_type_tuple(type_tuple):
            if isinstance(type_tuple, tuple):
                return type_tuple
            else:
                return tuple(v.strip().strip('`') for v in type_tuple.split(':', 1))

        return cls(conv_type_tuple(a) for a in args)

    def validate_composite_values(self, value):
        if value is None and self.nullable:
            return value
        if self._struct_as_dict:
            if isinstance(value, tuple):
                fields = getattr(value, "_fields", None) or self.field_types.keys()
                value = OrderedDict(compat.izip(fields, value))
            if isinstance(value, dict):
                return OrderedDict(
                    (validate_value(k, string), validate_value(value[k], tp))
                    for k, tp in six.iteritems(self.field_types)
                )
        else:
            if isinstance(value, tuple):
                return self.namedtuple_type(
                    *(validate_value(v, t) for v, t in zip(value, self.field_types.values()))
                )
            elif isinstance(value, dict):
                list_val = [
                    validate_value(value.get(key), field_type)
                    for key, field_type in self.field_types.items()
                ]
                return self.namedtuple_type(*list_val)
        raise ValueError(
            'Struct data type requires `tuple` or `dict`, instead of %s' % type(value)
        )


class Json(DataType):
    _type_id = 12

    _max_length = 8 * 1024 * 1024  # 8M

    def can_implicit_cast(self, other):
        if isinstance(other, six.string_types):
            other = validate_data_type(other)

        if isinstance(other, (String, Binary)):
            return True
        return super(Json, self).can_implicit_cast(other)

    def validate_value(self, val, max_field_size=None):
        if val is None and self.nullable:
            return True
        max_field_size = max_field_size or self._max_length
        if len(val) > max_field_size:
            raise ValueError(
                "InvalidData: Length of string(%s) is more than %sM.'" %
                (val, max_field_size / (1024 ** 2))
            )
        if not isinstance(val, (six.string_types, list, dict, six.integer_types, float)):
            raise ValueError("InvalidData: cannot accept %r as json", val)
        return True

    def cast_value(self, value, data_type):
        self._can_cast_or_throw(value, data_type)

        if isinstance(data_type, String):
            return _json.loads(utils.to_text(value))
        return value


tinyint = Tinyint()
smallint = Smallint()
int_ = Int()
bigint = Bigint()
float_ = Float()
double = Double()
string = String()
datetime = Datetime()
boolean = Boolean()
binary = Binary()
timestamp = Timestamp()
timestamp_ntz = TimestampNTZ()
interval_day_time = IntervalDayTime()
interval_year_month = IntervalYearMonth()
date = Date()
json = Json()

_odps_primitive_data_types = dict(
    [(t.name, t) for t in (
        tinyint, smallint, int_, bigint, float_, double, string, datetime, date, boolean,
        binary, timestamp, timestamp_ntz, interval_day_time, interval_year_month, json,
    )]
)


_composite_handlers = dict(
    varchar=Varchar,
    char=Char,
    decimal=Decimal,
    array=Array,
    map=Map,
    struct=Struct,
)


def parse_composite_types(type_str, handlers=None):
    handlers = handlers or _composite_handlers

    def _create_composite_type(typ, *args):
        prefix = None
        if ':' in typ:
            prefix, typ = typ.split(':')
            prefix = prefix.strip().strip('`')
            typ = typ.strip()
        if typ not in handlers:
            raise ValueError('Composite type %s not supported.' % typ.upper())
        ctype = handlers[typ].parse_composite(args)

        if prefix is None:
            return ctype
        else:
            return prefix, ctype

    token_stack = []
    bracket_stack = []
    token_start = 0
    type_str = type_str.strip()
    quoted = False

    for idx, ch in enumerate(type_str):
        if ch == '`':
            quoted = not quoted
        elif not quoted:
            if ch == '<' or ch == '(':
                bracket_stack.append(len(token_stack))
                token = type_str[token_start:idx].strip()
                token_stack.append(token)
                token_start = idx + 1
            elif ch == '>' or ch == ')':
                token = type_str[token_start:idx].strip()
                if token:
                    token_stack.append(token)
                bracket_pos = bracket_stack.pop()
                ctype = _create_composite_type(*token_stack[bracket_pos:])
                token_stack = token_stack[:bracket_pos]
                token_stack.append(ctype)
                token_start = idx + 1
            elif ch == ',':
                token = type_str[token_start:idx].strip()
                if token:
                    token_stack.append(token)
                token_start = idx + 1
    if len(token_stack) != 1:
        return _create_composite_type(type_str)
    return token_stack[0]


def validate_data_type(data_type):
    if isinstance(data_type, DataType):
        return data_type

    composite_err_msg = None
    if isinstance(data_type, six.string_types):
        data_type = data_type.strip().lower()
        if data_type in _odps_primitive_data_types:
            return _odps_primitive_data_types[data_type]

        try:
            return parse_composite_types(data_type)
        except ValueError as ex:
            composite_err_msg = str(ex)

    if composite_err_msg is not None:
        raise ValueError('Invalid data type: %s. %s' % (repr(data_type), composite_err_msg))
    raise ValueError('Invalid data type: %s' % repr(data_type))


integer_builtins = six.integer_types
float_builtins = (float,)
try:
    import numpy as np
    integer_builtins += (np.integer,)
    float_builtins += (np.float_,)
except ImportError:
    pass

_odps_primitive_to_builtin_types = OrderedDict((
    (bigint, integer_builtins),
    (tinyint, integer_builtins),
    (smallint, integer_builtins),
    (int_, integer_builtins),
    (double, float_builtins),
    (float_, float_builtins),
    (string, (six.text_type, six.binary_type)),
    (binary, six.binary_type),
    (datetime, _datetime),
    (boolean, bool),
    (interval_year_month, Monthdelta),
    (date, _date),
    (json, (list, dict, six.string_types, six.integer_types, float)),
))


integer_types = (tinyint, smallint, int_, bigint)


def infer_primitive_data_type(value):
    for data_type, builtin_types in six.iteritems(_odps_primitive_to_builtin_types):
        if isinstance(value, builtin_types):
            return data_type


def _patch_pd_types(data_type):
    if (
        timestamp not in _odps_primitive_to_builtin_types
        or timestamp_ntz not in _odps_primitive_to_builtin_types
        or interval_day_time not in _odps_primitive_to_builtin_types
    ) and isinstance(data_type, (BaseTimestamp, IntervalDayTime)):
        try:
            import pandas as pd
            _odps_primitive_to_builtin_types[timestamp] = pd.Timestamp
            _odps_primitive_to_builtin_types[timestamp_ntz] = pd.Timestamp
            _odps_primitive_to_builtin_types[interval_day_time] = pd.Timedelta
        except ImportError:
            raise ImportError(
                'To use %s in pyodps, you need to install pandas.', data_type.name.upper()
            )


def _validate_primitive_value(value, data_type):
    if value is None or type(value) is pd_na_type:
        return None

    if options.tunnel.string_as_binary:
        if isinstance(value, six.text_type):
            value = value.encode('utf-8')
    else:
        if isinstance(value, (bytearray, six.binary_type)):
            value = value.decode('utf-8')

    builtin_types = _odps_primitive_to_builtin_types[data_type]
    if isinstance(value, builtin_types):
        return value

    inferred_data_type = infer_primitive_data_type(value)
    if inferred_data_type is None:
        raise ValueError(
            'Unknown value type, cannot infer from value: %s, type: %s' % (value, type(value)))

    return data_type.cast_value(value, inferred_data_type)


def validate_value(value, data_type, max_field_size=None):
    _patch_pd_types(data_type)

    if data_type in _odps_primitive_to_builtin_types:
        res = _validate_primitive_value(value, data_type)
    elif isinstance(data_type, CompositeMixin):
        res = data_type.validate_composite_values(value)
    else:
        raise ValueError('Unknown data type: %s' % data_type)

    data_type.validate_value(res, max_field_size=max_field_size)
    return res



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from mars.services.web.core import MarsRequestHandler
from tornado.web import HTTPError

logger = logging.getLogger(__name__)


class PyODPSHandler(MarsRequestHandler):
    def get(self):
        pass

    def post(self):
        action = self.get_argument("action", "write_log")
        if action == "write_log":
            self._handle_write_log()
        elif action == "terminate":
            self._handle_terminate()
        else:  # pragma: no cover
            raise HTTPError(400, "Invalid action %r" % action)

    def _handle_write_log(self):
        content = self.get_argument("content")
        level = self.get_argument("level", "warning").lower()
        getattr(logger, level)(content)

    def _handle_terminate(self):
        from .cupid_service import CupidServiceClient
        from cupid import ContainerStatus, WorkItemProgress

        message = self.get_argument("message", "")
        CupidServiceClient().report_container_status(
            ContainerStatus.TERMINATED, message, WorkItemProgress.WIP_TERMINATING
        )


web_handlers = {
    "/api/pyodps": PyODPSHandler,
    "/api/pyodps_logger": PyODPSHandler,
}



# encoding: utf-8
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import uuid

from .core import DataFrame
from .. import options
from ..tempobj import register_temp_table
from ..examples import tables
from ..utils import TEMP_TABLE_PREFIX


def _base_example_loader(odps, method, **kwargs):
    roles = kwargs.pop('roles', None)

    table_name = TEMP_TABLE_PREFIX + 'df_' + method + '_' + str(uuid.uuid4()).lower().replace('-', '_')
    getattr(tables, method)(odps, table_name, **kwargs)
    register_temp_table(odps, table_name)
    odps.run_sql('alter table %s set lifecycle %d' % (table_name, options.temp_lifecycle))

    df = DataFrame(odps.get_table(table_name))
    if hasattr(df, 'roles') and roles:
        df = df.roles(**roles)
    return df


def create_ionosphere(odps, tunnel=None):
    """
    Create a PyODPS DataFrame with Johns Hopkins University ionosphere database.

    1. Title: Johns Hopkins University Ionosphere database

    2. Source Information:
       -- Donor: Vince Sigillito (vgs@aplcen.apl.jhu.edu)
       -- Date: 1989
       -- Source: Space Physics Group
                  Applied Physics Laboratory
                  Johns Hopkins University
                  Johns Hopkins Road
                  Laurel, MD 20723

    3. Relevant Information:
       This radar data was collected by a system in Goose Bay, Labrador.  This
       system consists of a phased array of 16 high-frequency antennas with a
       total transmitted power on the order of 6.4 kilowatts.  See the paper
       for more details.  The targets were free electrons in the ionosphere.
       "Good" radar returns are those showing evidence of some type of structure
       in the ionosphere.  "Bad" returns are those that do not; their signals pass
       through the ionosphere.

       Received signals were processed using an autocorrelation function whose
       arguments are the time of a pulse and the pulse number.  There were 17
       pulse numbers for the Goose Bay system.  Instances in this databse are
       described by 2 attributes per pulse number, corresponding to the complex
       values returned by the function resulting from the complex electromagnetic
       signal.

    4. Number of Instances: 351

    5. Number of Attributes: 34 plus the class attribute
       -- All 34 predictor attributes are continuous

    6. Attribute Information:
       -- All 34 are continuous, as described above
       -- The 35th attribute is either "good" or "bad" according to the definition
          summarized above.  This is a binary classification task.

    7. Missing Values: None

    :param odps: ODPS Object Reference
    :param tunnel: ODPS Tunnel Object Reference
    :return: Generated DataFrame
    """
    return _base_example_loader(odps, 'create_ionosphere', tunnel=tunnel, roles={'label': 'class'})


def create_iris(odps, tunnel=None):
    """
    Create a PyODPS DataFrame with iris plants database.

    1. Title: Iris Plants Database

    2. Sources:
         (a) Creator: R.A. Fisher
         (b) Donor: Michael Marshall (MARSHALL%PLU@io.arc.nasa.gov)
         (c) Date: July, 1988

    3. Relevant Information:
       --- This is perhaps the best known database to be found in the pattern
           recognition literature.  Fisher's paper is a classic in the field
           and is referenced frequently to this day.  (See Duda & Hart, for
           example.)  The DataFrame contains 3 classes of 50 instances each,
           where each class refers to a type of iris plant.  One class is
           linearly separable from the other 2; the latter are NOT linearly
           separable from each other.
       --- Predicted attribute: class of iris plant.
       --- This is an exceedingly simple domain.

    4. Number of Instances: 150 (50 in each of three classes)

    5. Number of Attributes: 4 numeric, predictive attributes and the class

    6. Attribute Information:
       1. sepal length in cm
       2. sepal width in cm
       3. petal length in cm
       4. petal width in cm
       5. class:
          -- Iris-setosa
          -- Iris-versicolour
          -- Iris-virginica

    7. Missing Attribute Values: None

    Summary Statistics:

    =============== ===== ===== ===== ===== ==================
                    Min   Max   Mean  SD    Class Correlation
    =============== ===== ===== ===== ===== ==================
    sepal length    4.3   7.9   5.84  0.83  0.7826
    sepal width     2.0   4.4   3.05  0.43  -0.4194
    petal length    1.0   6.9   3.76  1.76  0.9490  (high!)
    petal width     0.1   2.5   1.20  0.76  0.9565  (high!)
    =============== ===== ===== ===== ===== ==================

    8. Class Distribution: 33.3% for each of 3 classes.

    :param odps: ODPS Object Reference
    :param tunnel: ODPS Tunnel Object Reference
    :return: Generated DataFrame
    """
    return _base_example_loader(odps, 'create_iris', tunnel=tunnel, roles={'label': 'category'})



# Protocol Buffers - Google's data interchange format
# Copyright 2008 Google Inc.
# http://code.google.com/p/protobuf/
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This implementation is modified from google's original Protobuf implementation.
The original author is: robinson@google.com (Will Robinson).
Modified by onesuperclark@gmail.com(onesuper).
"""

import struct

from . import errors

TAG_TYPE_BITS = 3  # Number of bits used to hold type info in a proto tag.
_TAG_TYPE_MASK = (1 << TAG_TYPE_BITS) - 1  # 0x7

# These numbers identify the wire type of a protocol buffer value.
# We use the least-significant TAG_TYPE_BITS bits of the varint-encoded
# tag-and-type to store one of these WIRETYPE_* constants.
# These values must match WireType enum in //net/proto2/public/wire_format.h.
WIRETYPE_VARINT = 0
WIRETYPE_FIXED64 = 1
WIRETYPE_LENGTH_DELIMITED = 2
WIRETYPE_START_GROUP = 3
WIRETYPE_END_GROUP = 4
WIRETYPE_FIXED32 = 5
_WIRETYPE_MAX = 5

# Bounds for various integer types.
INT32_MAX = int((1 << 31) - 1)
INT32_MIN = int(-(1 << 31))
UINT32_MAX = (1 << 32) - 1

INT64_MAX = (1 << 63) - 1
INT64_MIN = -(1 << 63)
UINT64_MAX = (1 << 64) - 1

# "struct" format strings that will encode/decode the specified formats.
FORMAT_UINT32_LITTLE_ENDIAN = '<I'
FORMAT_UINT64_LITTLE_ENDIAN = '<Q'

# We'll have to provide alternate implementations of AppendLittleEndian*() on
# any architectures where these checks fail.
if struct.calcsize(FORMAT_UINT32_LITTLE_ENDIAN) != 4:
    raise AssertionError('Format "I" is not a 32-bit number.')
if struct.calcsize(FORMAT_UINT64_LITTLE_ENDIAN) != 8:
    raise AssertionError('Format "Q" is not a 64-bit number.')


def pack_tag(field_number, wire_type):
    """Returns an unsigned 32-bit integer that encodes the field number and
    wire type information in standard protocol message wire format.

    Args:
      field_number: Expected to be an integer in the range [1, 1 << 29)
      wire_type: One of the WIRETYPE_* constants.
    """
    if not 0 <= wire_type <= _WIRETYPE_MAX:
        raise errors.EncodeError('Unknown wire type: %d' % wire_type)
    return (field_number << TAG_TYPE_BITS) | wire_type


def unpack_tag(tag):
    """The inverse of PackTag().  Given an unsigned 32-bit number,
    returns a (field_number, wire_type) tuple.
    """
    return (tag >> TAG_TYPE_BITS), (tag & _TAG_TYPE_MASK)


def zig_zag_encode(value):
    """ZigZag Transform:  Encodes signed integers so that they can be
    effectively used with varint encoding.  See wire_format.h for
    more details.
    """
    if value >= 0:
        return value << 1
    return ((value << 1) ^ (~0)) | 0x1


def zig_zag_decode(value):
    """Inverse of ZigZagEncode()."""
    if not value & 0x1:
        return value >> 1
    return (value >> 1) ^ (~0)


# The *ByteSize() functions below return the number of bytes required to
# serialize "field number + type" information and then serialize the value.


def int32_byte_size(field_number, int32):
    return int64_byte_size(field_number, int32)


def int64_byte_size(field_number, int64):
    # Have to convert to uint before calling UInt64ByteSize().
    return UInt64ByteSize(field_number, 0xffffffffffffffff & int64)


def uint32_byte_size(field_number, uint32):
    return UInt64ByteSize(field_number, uint32)


def UInt64ByteSize(field_number, uint64):
    return _tag_byte_size(field_number) + _var_uint64_byte_size_no_tag(uint64)


def sint32_byte_size(field_number, int32):
    return uint32_byte_size(field_number, zig_zag_encode(int32))


def sint64_byte_size(field_number, int64):
    return UInt64ByteSize(field_number, zig_zag_encode(int64))


def fixed32_byte_size(field_number, fixed32):
    return _tag_byte_size(field_number) + 4


def fixed64_byte_size(field_number, fixed64):
    return _tag_byte_size(field_number) + 8


def sfixed32_byte_size(field_number, sfixed32):
    return _tag_byte_size(field_number) + 4


def sfixed64_byte_size(field_number, sfixed64):
    return _tag_byte_size(field_number) + 8


def float_byte_size(field_number, flt):
    return _tag_byte_size(field_number) + 4


def double_byte_size(field_number, double):
    return _tag_byte_size(field_number) + 8


def bool_byte_size(field_number, b):
    return _tag_byte_size(field_number) + 1


def enum_byte_size(field_number, enum):
    return uint32_byte_size(field_number, enum)


def string_byte_size(field_number, string):
    return (_tag_byte_size(field_number)
            + _var_uint64_byte_size_no_tag(len(string))
            + len(string))


def bytes_byte_size(field_number, b):
    return string_byte_size(field_number, b)


def group_byte_size(field_number, message):
    return (2 * _tag_byte_size(field_number)  # START and END group.
            + message.ByteSize())


def message_byte_size(field_number, message):
    return (_tag_byte_size(field_number)
            + _var_uint64_byte_size_no_tag(message.ByteSize())
            + message.ByteSize())


def message_set_item_byte_size(field_number, msg):
    # First compute the sizes of the tags.
    # There are 2 tags for the beginning and ending of the repeated group, that
    # is field number 1, one with field number 2 (type_id) and one with field
    # number 3 (message).
    total_size = (2 * _tag_byte_size(1) + _tag_byte_size(2) + _tag_byte_size(3))

    # Add the number of bytes for type_id.
    total_size += _var_uint64_byte_size_no_tag(field_number)

    message_size = msg.ByteSize()

    # The number of bytes for encoding the length of the message.
    total_size += _var_uint64_byte_size_no_tag(message_size)

    # The size of the message.
    total_size += message_size
    return total_size


# Private helper functions for the *ByteSize() functions above.


def _tag_byte_size(field_number):
    """Returns the bytes required to serialize a tag with this field number."""
    # Just pass in type 0, since the type won't affect the tag+type size.
    return _var_uint64_byte_size_no_tag(pack_tag(field_number, 0))


def _var_uint64_byte_size_no_tag(uint64):
    """Returns the bytes required to serialize a single varint.
    uint64 must be unsigned.
    """
    if uint64 > UINT64_MAX:
        raise errors.EncodeError('Value out of range: %d' % uint64)
    bytes = 1
    while uint64 > 0x7f:
        bytes += 1
        uint64 >>= 7
    return bytes



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools
import logging
from contextlib import contextmanager

from ....errors import ODPSError, NoPermission
from ....types import PartitionSpec
from ....compat import izip
from ....models import Table
from ....tunnel.tabletunnel import TableDownloadSession
from ...expr.arithmetic import And, Equal
from ...expr.reduction import *
from ...expr.element import Func
from ...utils import is_source_collection, is_source_partition
from ..frame import ResultFrame
from . import types

logger = logging.getLogger(__name__)


@contextmanager
def _open_reader(t, **kwargs):
    try:
        reader = t.open_reader(**kwargs)
        if reader.status != TableDownloadSession.Status.Normal:
            raise ODPSError('Intentionally reusing')
    except ODPSError:
        reader = t.open_reader(reopen=True, **kwargs)
    yield reader


class TunnelEngine(object):
    def __init__(self, odps):
        self._odps = odps

    @classmethod
    def _is_source_column(cls, expr, table):
        if not isinstance(expr, Column):
            return False

        odps_schema = table.table_schema
        if odps_schema.is_partition(expr.source_name):
            return False

        return True

    @classmethod
    def _can_propagate(cls, collection, no_filter=False, no_projection=False):
        if not isinstance(collection, CollectionExpr):
            return False

        if is_source_collection(collection) and \
                isinstance(collection._source_data, Table):
            return True
        if isinstance(collection, FilterPartitionCollectionExpr):
            return True
        if not no_filter and cls._filter_on_partition(collection):
            return True
        if not no_projection and cls._projection_on_source(collection):
            return True
        return False

    @classmethod
    def _projection_on_source(cls, expr):
        cols = []

        if isinstance(expr, ProjectCollectionExpr) and \
                cls._can_propagate(expr.input, no_projection=True):
            for col in expr.fields:
                source = next(expr.data_source())
                if not cls._is_source_column(col, source):
                    return False
                cols.append(col.source_name)
            return cols
        elif isinstance(expr, FilterPartitionCollectionExpr):
            return expr.schema.names
        return False

    @classmethod
    def _filter_on_partition(cls, expr):
        if not isinstance(expr, (FilterCollectionExpr, FilterPartitionCollectionExpr)) or \
                not cls._can_propagate(expr.input, no_filter=True):
            return False

        cols = []
        values = []

        def extract(expr):
            if isinstance(expr, Column):
                if is_source_partition(expr, next(expr.data_source())):
                    cols.append(expr.source_name)
                    return True
                else:
                    return False

            if isinstance(expr, And):
                for child in expr.args:
                    if not extract(child):
                        return False
            elif isinstance(expr, Equal) and isinstance(expr._rhs, Scalar) and \
                    not isinstance(expr._rhs, Func):  # skip the internal function
                if extract(expr._lhs):
                    values.append(expr._rhs.value)
                    return True
                else:
                    return False
            else:
                return False

            return True

        if not extract(expr.predicate):
            return False

        if len(cols) == len(values):
            return list(zip(cols, values))
        return False

    @classmethod
    def _to_partition_spec(cls, kv):
        spec = PartitionSpec()
        for k, v in kv:
            spec[k] = v
        return spec

    @classmethod
    def _partition_prefix(cls, all_partitions, filtered_partitions):
        filtered_partitions = sorted(six.iteritems(filtered_partitions.kv),
                                     key=lambda x: all_partitions.index(x[0]))
        if len(filtered_partitions) > len(all_partitions):
            return
        if not all(zip(l == r for l, r in zip(all_partitions, filtered_partitions))):
            return

        return cls._to_partition_spec(filtered_partitions)

    def execute(self, expr, ui=None, progress_proportion=1,
                head=None, tail=None, verify=False, update_progress_count=50):
        if isinstance(expr, (ProjectCollectionExpr, Summary)) and \
                len(expr.fields) == 1 and \
                isinstance(expr.fields[0], Count):
            expr = expr.fields[0]

        columns, partitions, count = (None, ) * 3
        if isinstance(expr, Count):
            if isinstance(expr.input, Column):
                input = expr.input.input
            else:
                input = expr.input
        elif isinstance(expr, SliceCollectionExpr):
            input = expr.input
        else:
            input = expr

        if verify:
            return self._can_propagate(input)
        if not self._can_propagate(input):
            return

        while True:
            if isinstance(input, FilterPartitionCollectionExpr):
                partition_kv = self._filter_on_partition(input)
                if not partition_kv:
                    return
                partitions = self._to_partition_spec(partition_kv)
                if not columns:
                    columns = self._projection_on_source(input)
                break
            else:
                ret = self._filter_on_partition(input)
                if ret:
                    partitions = self._to_partition_spec(ret)
                    input = input.input
                    continue

                ret = self._projection_on_source(input)
                if ret:
                    columns = ret
                    input = input.input
                    continue
                break

        table = next(expr.data_source())
        partition, filter_all_partitions = None, True
        if table.table_schema.partitions:
            if partitions is not None:
                partition = self._partition_prefix(
                    [p.name for p in table.table_schema.partitions], partitions)
                if partition is None:
                    return
                if len(table.table_schema.partitions) != len(partitions):
                    filter_all_partitions = False
            else:
                filter_all_partitions = False

        if isinstance(expr, Count):
            if not filter_all_partitions:
                # if not filter all partitions, fall back to ODPS SQL to calculate count
                return
            try:
                with _open_reader(table, partition=partition) as reader:
                    ui.inc(progress_proportion)
                    return reader.count
            except ODPSError:
                return
        else:
            logger.info(
                'Try fetching data from tunnel. If it takes a long time, please try running your code '
                'with distributed capabilities, see related section in '
                'https://www.alibabacloud.com/help/en/maxcompute/latest/platform-instructions-overview#section-wy0-8st-f60 '
                'for more details.'
            )
            ui.status('Try to download data with tunnel...', clear_keys=True)
            if isinstance(expr, SliceCollectionExpr):
                if expr.start:
                    raise ExpressionError('For ODPS backend, slice\'s start cannot be specified')
                count = expr.stop
            try:
                data = []

                start, size, step = None, None, None
                if head is not None:
                    size = min(head, count) if count is not None else head
                elif tail is not None:
                    if filter_all_partitions:
                        start = None if count is None else max(count - tail, 0)
                        size = tail if count is None else min(count, tail)
                    else:
                        # tail on multi partitions, just fall back to SQL
                        return
                else:
                    size = count

                fetch_partitions = [partition] if filter_all_partitions else \
                    (p.name for p in table.iterate_partitions(partition))
                if tail is not None:
                    fetch_partitions = list(fetch_partitions)[::-1]
                if size is None:
                    fetch_partitions = list(fetch_partitions)

                cum = 0
                last_percent = 0
                for curr_part, partition in izip(itertools.count(1), fetch_partitions):
                    rest = size - cum if size is not None else None
                    finished = False

                    with _open_reader(table, partition=partition) as reader:
                        if tail is not None and start is None:
                            s = max(reader.count - tail, 0)
                            start = s if start is None else max(s, start)

                        unique_columns = list(OrderedDict.fromkeys(columns)) if columns is not None else None
                        for i, r in izip(itertools.count(1),
                                         reader.read(start=start, count=rest, columns=unique_columns)):
                            if size is not None and cum > size - 1:
                                finished = True
                                break
                            cum += 1
                            if cum % update_progress_count == 0:
                                if size is not None:
                                    p = float(cum) / size * progress_proportion
                                    ui.inc(p - last_percent)
                                    last_percent = p
                                else:
                                    p = ((curr_part - 1) / len(fetch_partitions) +
                                         float(i) / reader.count / len(fetch_partitions)) * progress_proportion
                                    ui.inc(p - last_percent)
                                    last_percent = p
                            if partition:
                                spec = PartitionSpec(partition) if not isinstance(partition, PartitionSpec) \
                                    else partition
                                self._fill_back_partition_values(r, table, spec.kv)
                            if columns is None or len(unique_columns) == len(columns):
                                data.append(r.values)
                            else:
                                data.append([r[n] for n in columns])

                    if finished:
                        break

                if last_percent < progress_proportion:
                    ui.inc(progress_proportion - last_percent)
                return ResultFrame(data, schema=expr._schema)
            except NoPermission:
                raise
            except ODPSError:
                return

    @classmethod
    def _fill_back_partition_values(cls, record, table, pkv):
        if pkv:
            for k, v in six.iteritems(pkv):
                if k in record and record[k] is None:
                    # fill back the partition data which is lost in the tunnel
                    record[k] = types.odps_types.validate_value(
                        v, table.table_schema.get_type(k)
                    )



# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import io
import logging
import os
import tempfile
import time
import uuid
import socket
import sys

from mars.actors.pool.gevent_pool import ActorPool
from mars.deploy.kubernetes.core import K8SPodsIPWatcher

logger = logging.getLogger(__name__)


class CupidK8SPodsIPWatcher(K8SPodsIPWatcher):
    _pod_to_port = dict()

    @property
    def cupid_kv(self):
        if not hasattr(self, "_cupid_kv"):
            from cupid import context

            self._cupid_kv = context().kv_store()
        return self._cupid_kv

    def _extract_pod_name_ep(self, pod_data):
        from cupid.runtime import RuntimeContext

        pod_name = pod_data["metadata"]["name"]
        if not RuntimeContext.is_context_ready():
            logger.debug("Cupid context not ready, pod name: {}".format(pod_name))
            return pod_name, None

        if self._pod_to_port.get(pod_name):
            pod_port = self._pod_to_port[pod_name]
        else:
            pod_kv_data = self.cupid_kv.get(pod_name)
            if pod_kv_data:
                pod_port = self._pod_to_port[pod_name] = json.loads(pod_kv_data)[
                    "endpoint"
                ].rsplit(":", 1)[-1]
                logger.debug(
                    "Get port from kvstore, name: {}, port: {}".format(
                        pod_name, pod_port
                    )
                )
            else:
                pod_port = None
                if pod_name not in self._pod_to_port:
                    logger.debug(
                        "Cannot get port from kvstore, name: {}".format(pod_name)
                    )
                    self._pod_to_port[pod_name] = None
        pod_endpoint = "%s:%s" % (pod_data["status"]["pod_ip"], pod_port)
        return pod_name, pod_endpoint if pod_port else None

    def _get_pod_to_ep(self, service_type=None):
        if service_type is not None:
            query = (
                self._pool.spawn(
                    self._client.list_namespaced_pod,
                    namespace=self._k8s_namespace,
                    label_selector=self._get_label_selector(service_type),
                )
                .result()
                .to_dict()
            )
        else:
            query = (
                self._pool.spawn(
                    self._client.list_namespaced_pod,
                    namespace=self._k8s_namespace,
                    label_selector=self._get_label_selector(),
                )
                .result()
                .to_dict()
            )
        result = dict()
        for el in query["items"]:
            name, pod_ep = self._extract_pod_name_ep(el)
            if pod_ep is not None and not self._extract_pod_ready(el):
                pod_ep = None
            if el["status"]["phase"] == "Running":
                result[name] = pod_ep
        return result

    @staticmethod
    def _extract_pod_ready(obj_data):
        if obj_data["status"]["phase"] != "Running":
            return False
        return True


class CupidActorPool(ActorPool):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._cupid_context = None
        self._env_path = tempfile.mkdtemp(prefix="mars-pool-")
        self._channel_file = dict()

    def _prepare_process_channel(self, idx):
        envs = self._cupid_context.prepare_channel()
        envs_dict = {env.name: env.value for env in envs}
        with open(self._channel_file[idx], "w") as env_file:
            env_file.write(json.dumps(envs_dict))

    def run(self):
        if self.processes:
            return

        super().run()

        from cupid import context

        self._cupid_context = context()

        kvstore = self._cupid_context.kv_store()
        advertise_endpoint = (
            self.advertise_address.split(":")[0] + ":" + self.address.split(":")[-1]
        )
        kvstore[os.environ["MARS_K8S_POD_NAME"]] = json.dumps(
            dict(endpoint=advertise_endpoint)
        )
        logger.debug(
            "Endpoint %s written to %s",
            advertise_endpoint,
            os.environ["MARS_K8S_POD_NAME"],
        )

        for idx in range(len(self.processes)):
            self._prepare_process_channel(idx)

    def stop(self):
        import shutil

        shutil.rmtree(self._env_path)
        if self._cupid_context is not None:
            self._cupid_context.channel_client.stop()
        super().stop()

    def pre_process_start(self, idx):
        self._channel_file[idx] = os.path.join(self._env_path, str(uuid.uuid4()))

    def post_process_start_child(self, idx):
        try:
            # Patch import here.
            # The reason is that tensorflow relies on protobuf 3+,
            # meanwhile, cupid channel relies on protobuf 2.4,
            # however, when cupid channel started below,
            # tensorflow will recognize the old version of protobuf
            # even when we set LD_LIBRARY_PATH,
            # so we import tensorflow in advance to prevent from potential crash.
            import tensorflow  # noqa: F401
        except ImportError:
            pass

        # set STDOUT to unbuffer mode
        sys.stdout = io.TextIOWrapper(
            open(sys.stdout.fileno(), "wb", 0), write_through=True
        )

        while not os.path.exists(self._channel_file[idx]):
            time.sleep(1)
        try:
            with open(self._channel_file[idx], "r") as env_file:
                envs = json.loads(env_file.read())
        except:
            time.sleep(1)
            with open(self._channel_file[idx], "r") as env_file:
                envs = json.loads(env_file.read())

        from cupid import context

        os.environ.update(envs)
        _proc_cupid_context = context()  # noqa: F841
        odps_envs = {
            "ODPS_BEARER_TOKEN": os.environ["BEARER_TOKEN_INITIAL_VALUE"],
            "ODPS_ENDPOINT": os.environ["ODPS_RUNTIME_ENDPOINT"],
        }
        os.environ.update(odps_envs)
        logger.info("Started channel for process index %s.", idx)


class CupidServiceMixin:
    def create_scheduler_discoverer(self):
        try:
            self.scheduler_discoverer = CupidK8SPodsIPWatcher(
                label_selector="name=marsscheduler"
            )
        except TypeError:
            self.scheduler_discoverer = CupidK8SPodsIPWatcher()
        finally:
            os.environ.pop("KUBE_API_ADDRESS", None)

    def create_pool(self, *args, **kwargs):
        kwargs["pool_cls"] = CupidActorPool
        return super().create_pool(*args, **kwargs)

    def parse_args(self, parser, argv, environ=None):
        args = super().parse_args(parser, argv, environ)
        if os.environ.get("VM_ENGINE_TYPE") == "hyper":
            args.advertise = socket.gethostbyname(socket.gethostname())
        return args



# Copyright 1999-2024 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Restful client enhanced by URL building and request signing facilities.
"""
from __future__ import absolute_import
import json
import logging
import platform
import re
import threading
from string import Template

import requests
try:
    from requests import ConnectTimeout
except ImportError:
    from requests import Timeout as ConnectTimeout
try:
    import requests_unixsocket
except ImportError:
    requests_unixsocket = None

from . import __version__
from . import errors, utils
from .config import options
from .utils import get_package_version, get_survey_calls, clear_survey_calls
from .compat import six, urlparse

try:
    import requests.packages.urllib3.util.ssl_
    requests.packages.urllib3.util.ssl_.DEFAULT_CIPHERS = 'ALL'
    requests.packages.urllib3.disable_warnings()
except ImportError:
    pass

try:
    import urllib3.util.ssl_
    urllib3.util.ssl_.DEFAULT_CIPHERS = 'ALL'
    urllib3.disable_warnings()
except ImportError:
    pass


logger = logging.getLogger(__name__)

_default_user_agent = None

_v4_sign_fallback_msgs = [
    "need ak v3 support",
    "accesskey acl denied",
]


def default_user_agent():
    global _default_user_agent
    if _default_user_agent is not None:
        return _default_user_agent

    py_implementation = platform.python_implementation()
    py_version = platform.python_version()
    try:
        py_system = platform.system()
        py_release = platform.release()
    except IOError:
        py_system = 'Unknown'
        py_release = 'Unknown'

    ua_template = Template(
        options.user_agent_pattern
        or '$pyodps_version $mars_version $maxframe_version $python_version $os_version'
    )
    substitutes = dict(
        pyodps_version='%s/%s' % ('pyodps', __version__),
        python_version='%s/%s' % (py_implementation, py_version),
        os_version='%s/%s' % (py_system, py_release),
        mars_version='',
        maxframe_version='',
    )

    try:
        from mars import __version__ as mars_version
    except ImportError:
        mars_version = None
    if mars_version:
        substitutes["mars_version"] = '%s/%s' % ('mars', mars_version)

    try:
        maxframe_version = get_package_version("maxframe")
    except:
        maxframe_version = None
    if maxframe_version:
        substitutes["maxframe_version"] = '%s/%s' % ('maxframe', maxframe_version)

    _default_user_agent = ua_template.safe_substitute(**substitutes)
    _default_user_agent = re.sub(" +", " ", _default_user_agent).strip()

    try:
        from .internal.rest import get_internal_user_agent_suffix

        _default_user_agent += " " + get_internal_user_agent_suffix()
    except:
        pass
    return _default_user_agent


class RestClient(object):
    _session_local = threading.local()
    _endpoints_without_v4_sign = set()

    def __init__(
        self, account, endpoint, project=None, schema=None, user_agent=None, region_name=None, **kwargs
    ):
        if endpoint.endswith('/'):
            endpoint = endpoint[:-1]
        self._account = account
        self._endpoint = endpoint
        self._region_name = region_name
        self._user_agent = user_agent or default_user_agent()
        self.project = project
        self.schema = schema
        self._proxy = kwargs.get('proxy')
        self._app_account = kwargs.get('app_account')
        self._tag = kwargs.get('tag')
        if isinstance(self._proxy, six.string_types):
            self._proxy = dict(http=self._proxy, https=self._proxy)

    @property
    def endpoint(self):
        return self._endpoint

    @property
    def account(self):
        return self._account

    @property
    def app_account(self):
        return self._app_account

    @property
    def region_name(self):
        return self._region_name

    @property
    def session(self):
        try:
            session_cache = type(self)._session_local.session_cache
        except AttributeError:
            session_cache = type(self)._session_local.session_cache = dict()
        try:
            return session_cache[self._endpoint]
        except KeyError:
            pass

        parsed_url = urlparse(self._endpoint)
        adapter_options = dict(
            pool_connections=options.pool_connections,
            pool_maxsize=options.pool_maxsize,
            max_retries=options.retry_times,
        )
        if parsed_url.scheme == "http+unix":
            session = requests_unixsocket.Session()
            session.mount(
                'http+unix://',
                requests_unixsocket.adapters.UnixAdapter(**adapter_options)
            )
        else:
            session = requests.Session()
            # mount adapters with retry times
            session.mount(
                'http://', requests.adapters.HTTPAdapter(**adapter_options)
            )
            session.mount(
                'https://', requests.adapters.HTTPAdapter(**adapter_options)
            )
        session_cache[self._endpoint] = session
        return session

    def request(self, url, method, stream=False, **kwargs):
        sign_region_name = kwargs.get("region_name") or self._region_name
        if self._endpoint in self._endpoints_without_v4_sign or not options.enable_v4_sign:
            sign_region_name = None

        while True:
            kwargs["region_name"] = sign_region_name
            try:
                return self._request(url, method, stream=stream, **kwargs)
            except errors.InternalServerError as ex:
                ex_msg = str(ex).lower()
                if sign_region_name is None or all(
                    msg not in ex_msg for msg in _v4_sign_fallback_msgs
                ):
                    raise
                self._endpoints_without_v4_sign.add(self._endpoint)
                sign_region_name = None
            except errors.InvalidParameter as ex:
                if sign_region_name is None or "ODPS-0410051" not in str(ex):
                    # Invalid credentials error not received from server
                    raise
                self._endpoints_without_v4_sign.add(self._endpoint)
                sign_region_name = None
            except errors.AuthorizationRequired as ex:
                if sign_region_name is None or "invalid or missing" not in str(ex):
                    raise
                self._endpoints_without_v4_sign.add(self._endpoint)
                sign_region_name = None

    def _request(self, url, method, stream=False, **kwargs):
        self.upload_survey_log()

        region_name = kwargs.pop("region_name", None)

        logger.debug('Start request.')
        logger.debug('%s: %s', method.upper(), url)
        if logger.getEffectiveLevel() <= logging.DEBUG:
            for k, v in kwargs.items():
                logger.debug('%s: %s', k, v)

        # Construct user agent without handling the letter case.
        headers = kwargs.get('headers', {})
        headers = {k: str(v) for k, v in six.iteritems(headers)}
        headers['User-Agent'] = self._user_agent
        kwargs['headers'] = headers
        params = kwargs.setdefault('params', {})

        actions = kwargs.pop("actions", None) or kwargs.pop("action", None) or []
        if isinstance(actions, six.string_types):
            actions = [actions]
        if actions:
            separator = "?" if "?" not in url else "&"
            url += separator + "&".join(actions)

        curr_project = kwargs.pop("curr_project", None) or self.project
        if 'curr_project' not in params and curr_project is not None:
            params['curr_project'] = curr_project

        curr_schema = kwargs.pop("curr_schema", None) or self.schema
        if 'curr_schema' not in params and curr_schema is not None:
            params['curr_schema'] = curr_schema

        timeout = kwargs.pop('timeout', None)
        req = requests.Request(method, url, **kwargs)
        prepared_req = req.prepare()
        logger.debug("request url + params %s", prepared_req.path_url)

        prepared_req.headers.pop('Authorization', None)
        prepared_req.headers.pop('application-authentication', None)
        self._account.sign_request(prepared_req, self._endpoint, region_name=region_name)
        if getattr(self, '_app_account', None) is not None:
            self._app_account.sign_request(
                prepared_req, self._endpoint, region_name=region_name
            )

        try:
            res = self.session.send(
                prepared_req,
                stream=stream,
                timeout=timeout or (options.connect_timeout, options.read_timeout),
                verify=options.verify_ssl,
                proxies=self._proxy,
            )
        except ConnectTimeout:
            raise errors.ConnectTimeout('Connecting to endpoint %s timeout.' % self._endpoint)

        logger.debug('response.status_code %d', res.status_code)
        logger.debug('response.headers: \n%s', res.headers)
        if not stream:
            logger.debug('response.content: %s\n', res.content)
        # Automatically detect error
        if not self.is_ok(res):
            errors.throw_if_parsable(res, self._endpoint, self._tag)
        return res

    def get(self, url, stream=False, **kwargs):
        return self.request(url, 'get', stream=stream, **kwargs)

    def post(self, url, data=None, **kwargs):
        data = utils.to_binary(data, encoding='utf-8') if isinstance(data, six.string_types) else data
        return self.request(url, 'post', data=data, **kwargs)

    def put(self, url, data=None, **kwargs):
        data = utils.to_binary(data) if isinstance(data, six.string_types) else data
        return self.request(url, 'put', data=data, **kwargs)

    def head(self, url, **kwargs):
        return self.request(url, 'head', **kwargs)

    def delete(self, url, **kwargs):
        return self.request(url, 'delete', **kwargs)

    def upload_survey_log(self):
        try:
            from .models.core import RestModel

            survey = get_survey_calls()
            clear_survey_calls()
            if not survey:
                return
            if self.project is None:
                return
            url = '/'.join([self.endpoint, 'projects', RestModel._encode(self.project), 'logs'])
            self.put(url, json.dumps(survey))
        except:
            pass

    # Misc helper methods
    def is_ok(self, resp):
        return resp.ok



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import uuid
import warnings

from .. import op, exporters
from ..core import AlgoExprMixin, AlgoCollectionExpr
from ...utils import MLField, FieldRole, FieldContinuity, import_class_member, build_model_table_name, ML_ARG_PREFIX
from ....compat import six
from ....config import options
from ....errors import NoSuchObject
from ....df.expr.collections import CollectionExpr, Expr
from ....df.expr.dynamic import DynamicMixin


_PREDICTION_ARGS = """
inputTableName modelName outputTableName featureColNames appendColNames inputTablePartitions
resultColName scoreColName detailColName enableSparse itemDelimiter kvDelimiter
""".split()
_PREDICTION_EXPORTERS = {
    "inputTableName": lambda e: exporters.get_input_table_name(e, "input"),
    "inputTablePartitions": lambda e: exporters.get_input_partitions(e, "input"),
    "modelName": lambda e: exporters.get_input_model_name(e, "model"),
    "outputTableName": lambda e: exporters.get_output_table_name(e, "output"),
    "outputTablePartition": lambda e: exporters.get_output_table_partitions(e, "output"),
    "appendColNames": lambda e: exporters.get_original_columns(e, "input"),
    "enableSparse": lambda e: exporters.get_enable_sparse(e, 'enableSparse', 'input'),
    "itemDelimiter": lambda e: exporters.get_item_delimiter(e, 'itemDelimiter', 'input'),
    "kvDelimiter": lambda e: exporters.get_kv_delimiter(e, 'kvDelimiter', 'input'),
    "featureColNames": lambda e: exporters.get_sparse_predict_feature_columns(e, 'featureColNames', 'input'),
}
_PREDICTION_META = {
    'xflowName': 'prediction',
    'xflowProjectName': 'algo_public',
}


class PredictionCollectionExpr(AlgoCollectionExpr):
    node_name = 'Prediction'
    _algo = 'prediction'
    _args = '_mlattr_input', '_mlattr_model'
    _exported = _PREDICTION_ARGS
    _exporters = _PREDICTION_EXPORTERS
    algo_meta = _PREDICTION_META

    def _init(self, *args, **kwargs):
        super(PredictionCollectionExpr, self)._init(*args, **kwargs)

        self._init_attr('_params', dict())
        self._init_attr('_engine_kw', dict())

    @property
    def input(self):
        return self._mlattr_input

    @property
    def model(self):
        return self._mlattr_model


class ModelDataCollectionExpr(CollectionExpr):
    node_name = 'ModelData'
    __slots__ = '_data_item',
    _args = '_mlattr_model',

    def _init(self, *args, **kwargs):
        kwargs.pop('register_expr', None)
        model_inputs = kwargs.pop('model_inputs', [])
        for mi in model_inputs:
            kwargs.pop(ML_ARG_PREFIX + mi, None)

        super(ModelDataCollectionExpr, self)._init(*args, **kwargs)

    @property
    def model(self):
        return self._mlattr_model

    def table_name(self, model_name=None):
        return self.model.table_name(self._data_item, model_name)

    def convert_params(self, src_expr=None):
        if src_expr is not None:
            src_expr = src_expr.model
        return self.model.convert_params(src_expr=src_expr)

    def accept(self, visitor):
        if self._source_data is not None:
            super(ModelDataCollectionExpr, self).accept(visitor)


class ODPSModelExpr(AlgoExprMixin, Expr):
    __slots__ = '_source_data', '_is_offline_model', '_model_collections', \
                '_model_params', '_predict_fields', '_predictor', '_recommender', '_pmml'
    _suffix = 'ModelExpr'
    _non_table = True
    _pmml_members = 'load_pmml', 'segments', 'regression'
    node_name = 'Model'

    def _fill_attr(self, attr, val):
        if not hasattr(self, attr):
            setattr(self, attr, val)
        if getattr(self, attr) is None:
            setattr(self, attr, val)

    def _init(self, *args, **kwargs):
        register_expr = kwargs.pop('register_expr', False)

        p_args = [a.cache() if isinstance(a, Expr) else a for a in args]
        p_kw = dict((k, self.cache_input(v)) for k, v in six.iteritems(kwargs))

        super(ODPSModelExpr, self)._init(*p_args, **p_kw)
        self.cache(False)

        self._fill_attr('_model_collections', dict())
        self._fill_attr('_model_params', dict())
        self._fill_attr('_is_offline_model', True)

        if hasattr(self, 'algo_meta'):
            for fun_name in ['predictor', 'recommender']:
                self._init_attr('_' + fun_name, self.algo_meta.get(fun_name, None))

        self._fill_attr('_predict_fields', [
            MLField('prediction_result', 'string', FieldRole.PREDICTED_CLASS, continuity=FieldContinuity.DISCRETE,
                    is_append=True),
            MLField('prediction_score', 'double', [FieldRole.PREDICTED_SCORE, FieldRole.PREDICTED_VALUE],
                    continuity=FieldContinuity.CONTINUOUS, is_append=True),
            MLField('prediction_detail', 'string', FieldRole.PREDICTED_DETAIL, is_append=True),
        ])

        if register_expr:
            self.register_expr()

    def _predict_offline_model(self, expr):
        df = PredictionCollectionExpr(register_expr=True, _mlattr_input=expr, _mlattr_model=self,
                                      _schema=expr.schema, _exec_id=uuid.uuid4(), _output_name='output')
        df._ml_uplink = [expr]

        append_op = op.StaticFieldChangeOperation(self._predict_fields, is_append=True)
        df._perform_operation(append_op)
        df._rebuild_df_schema(isinstance(expr, DynamicMixin))
        return df

    def _predict_tables_model(self, *args, **kwargs):
        from ...algolib.loader import dispatch_args
        method_path = kwargs.pop('_method_path', self._predictor)
        _predictor = import_class_member(method_path)
        members = set(p for p in dir(_predictor) if not p.startswith('_') and
                      not p.startswith('set_') and not callable(getattr(_predictor, p)))
        ml_args, p_args, ml_kw, p_kw = dispatch_args(members, *args, **kwargs)
        return _predictor(*p_args, **p_kw).transform(self, *ml_args, **ml_kw)

    def predict(self, *args, **kwargs):
        if self._is_offline_model:
            return self._predict_offline_model(*args, **kwargs)
        else:
            return self._predict_tables_model(*args, **kwargs)

    def _recommend(self, *args, **kwargs):
        kw = kwargs.copy()
        kw['_method_path'] = self._recommender
        return self._predict_tables_model(*args, **kw)

    def __getattr__(self, item):
        def raise_err():
            raise AttributeError("'{0}' object has no attribute '{1}'".format(type(self).__name__, item))

        if item.startswith('_'):
            raise_err()
        if item == 'recommend' and getattr(self, '_recommender', None) is not None:
            return self._recommend
        if item in type(self)._pmml_members:
            if not self._is_offline_model:
                raise_err()
            warnings.warn('Direct methods on Pmml Models is deprecated, see documentation '
                          'for more details.', DeprecationWarning)
            return getattr(self, '_' + item)
        if self._is_offline_model or item not in self._model_collections:
            raise_err()
        if item in self._model_collections:
            return self._model_collections[item]
        raise_err()

    def _data_source(self):
        if hasattr(self, '_source_data') and self._source_data is not None:
            yield self._source_data

    def table_name(self, item_name, model_name=None):
        model_name = model_name or str(self._source_data.name)
        return build_model_table_name(model_name, item_name)

    def accept(self, visitor):
        visitor.visit_algo(self)

    def get_cached(self, data):
        if self._is_offline_model:
            try:
                if not options.ml.dry_run:
                    data.reload()
            except NoSuchObject:
                from ....df.backends.context import context
                context.uncache(self)
                return None
            mod = type(self)(_source_data=data)
        else:
            mod = ODPSModelExpr(_source_data=data, _is_offline_model=False, _model_params=data.params.copy(),
                                _predictor=data.params.get('predictor'),
                                _recommender=data.params.get('recommender'))
            data_exprs = dict()
            for k, v in six.iteritems(data.tables):
                data_exprs[k] = ModelDataCollectionExpr(_mlattr_model=mod, _data_item=k)
                data_exprs[k]._source_data = v
            mod._model_collections = data_exprs
        mod._need_cache = False
        return mod

    def _load_pmml(self):
        if not getattr(self, '_pmml', None):
            pmml_obj = self.execute()
            self._pmml = pmml_obj.pmml
        return self._pmml

    @property
    def _regression(self):
        from .pmml import PmmlResult, PmmlRegressionResult
        self._load_pmml()
        result = PmmlResult(self._pmml)
        if not isinstance(result, PmmlRegressionResult):
            raise ValueError('`RegressionModel` element not found in PMML.')
        return result

    @property
    def _segments(self):
        from .pmml import PmmlResult, PmmlSegmentsResult
        self._load_pmml()
        result = PmmlResult(self._pmml)
        if not isinstance(result, PmmlSegmentsResult):
            raise ValueError('`Segmentation` element not found in PMML.')
        return result

    def persist(self, name, project=None, drop_model=False, **kwargs):
        """
        Persist the execution into a new model.

        :param name: model name
        :param project: name of the project
        :param drop_model: drop model before creation
        """
        return super(ODPSModelExpr, self).persist(name, project=project, drop_model=drop_model, **kwargs)

    def verify(self):
        return super(ODPSModelExpr, self).verify()


class PmmlModel(ODPSModelExpr):
    def _init(self, *args, **kwargs):
        if len(args) == 1:
            model = args[0]
        else:
            model = kwargs.pop('_source_data', None)
            if model is None:
                raise ValueError('ODPS offline model should be provided.')
        kwargs['_source_data'] = model
        kwargs['_is_offline_model'] = True
        kwargs['_exec_id'] = str(uuid.uuid4())

        super(PmmlModel, self)._init(*args, **kwargs)

        self.executed = True

    def predict(self, *args, **kwargs):
        """
        Predict given DataFrame using the given model. Actual prediction steps will not
        be executed till an operational step is called.

        After execution, three columns will be appended to the table:

        +-------------------+--------+----------------------------------------------------+
        | Field name        | Type   | Comments                                           |
        +===================+========+====================================================+
        | prediction_result | string | field indicating the predicted label, absent if    |
        |                   |        | the model is a regression model                    |
        +-------------------+--------+----------------------------------------------------+
        | prediction_score  | double | field indicating the score value if the model is   |
        |                   |        | a classification model, or the predicted value if  |
        |                   |        | the model is a regression model.                   |
        +-------------------+--------+----------------------------------------------------+
        | prediction_detail | string | field in JSON format indicating the score for      |
        |                   |        | every class.                                       |
        +-------------------+--------+----------------------------------------------------+

        :type df: DataFrame
        :rtype: DataFrame

        :Example:

        >>> model = PmmlModel(odps.get_offline_model('model_name'))
        >>> data = DataFrame(odps.get_table('table_name'))
        >>> # prediction below will not be executed till predicted.persist is called
        >>> predicted = model.predict(data)
        >>> predicted.persist('predicted')
        """
        return super(PmmlModel, self).predict(*args, **kwargs)


class TablesModel(ODPSModelExpr):
    def _init(self, *args, **kwargs):
        from ....df import DataFrame

        if len(args) == 1:
            model = args[0]
        else:
            model = kwargs.pop('_source_data', None)
            if model is None:
                raise ValueError('Name of the TablesModel should be provided.')

        model_collection = dict()
        for name, tb in six.iteritems(model.tables):
            model_collection[name] = DataFrame(tb)

        kwargs['_source_data'] = model
        kwargs['_is_offline_model'] = False
        kwargs['_model_collections'] = model_collection
        kwargs['_model_params'] = model.params

        for k in ('predictor', 'recommender'):
            kwargs['_' + k] = model.params.get(k)

        kwargs['_exec_id'] = str(uuid.uuid4())

        super(TablesModel, self)._init(*args, **kwargs)

    def predict(self, *args, **kwargs):
        """
        Predict given DataFrame using the given model. Actual prediction steps will not
        be executed till an operational step is called.

        :param list[DataFrame] args: input DataFrames to be predicted
        :param kwargs: named input DataFrames or prediction parameters, details can be found in ''Predictor Parameters'' section of training algorithms.

        A :class:`DataFrame` object will be generated for input data. Output fields may be referenced in the
        documents of training algorithms.
        """
        return super(TablesModel, self).predict(*args, **kwargs)


class TablesModelResult(object):
    def __init__(self, params, results):
        self._params = params
        self._results = results

    @property
    def params(self):
        return self._params

    @property
    def results(self):
        return self._results

    def __getattr__(self, item):
        if item in self._results:
            return self._results[item]
        raise AttributeError



# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
from ._version import __version__

__all__ = ['ODPS', 'DataFrame', 'options']

if sys.version_info[0] == 2 and sys.version_info[:2] < (2, 7):
    raise Exception('pyodps supports python 2.7+ (including python 3+).')

from .config import options
from .core import ODPS
from .df import DataFrame, Scalar, RandomScalar, NullScalar
from .inter import setup, enter, teardown, list_rooms
from .utils import show_versions, write_log as log
try:
    from .ipython import load_ipython_extension
except ImportError:
    pass
try:
    from sqlalchemy.dialects import registry

    registry.register('odps', 'odps.sqlalchemy_odps', 'ODPSDialect')
except ImportError:
    pass


def install_plugins():
    try:
        from .ml import install_plugin, install_mixin
    except (ImportError, SyntaxError):
        pass


install_plugins()
del install_plugins



# Protocol Buffers - Google's data interchange format
# Copyright 2008 Google Inc.
# http://code.google.com/p/protobuf/
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This implementation is modified from google's original Protobuf implementation.
The original author is: robinson@google.com (Will Robinson).
Modified by onesuperclark@gmail.com(onesuper).
"""

import struct

from . import errors
from . import wire_format


class InputStream(object):
    """Contains all logic for reading bits, and dealing with stream position.

    If an InputStream method ever raises an exception, the stream is left
    in an indeterminate state and is not safe for further use.
    """

    def __init__(self, input):
        self._input = input
        self._pos = 0

    def position(self):
        """Returns the current position in the stream, or equivalently, the
        number of bytes read so far.
        """
        return self._pos

    def read_string(self, size):
        """Reads up to 'size' bytes from the stream, stopping early
        only if we reach the end of the stream.  Returns the bytes read
        as a string.
        """
        if size < 0:
            raise errors.DecodeError('Negative size %d' % size)
        s = self._input.read(size)
        if len(s) != size:
            raise errors.DecodeError('String claims to have %d bytes, but read %d' % (size, len(s)))
        self._pos += len(s)  # Only advance by the number of bytes actually read.
        return s

    def read_little_endian32(self):
        """Interprets the next 4 bytes of the stream as a little-endian
        encoded, unsiged 32-bit integer, and returns that integer.
        """
        try:
            i = struct.unpack(wire_format.FORMAT_UINT32_LITTLE_ENDIAN,
                              self._input.read(4))
            self._pos += 4
            return i[0]  # unpack() result is a 1-element tuple.
        except struct.error as e:
            raise errors.DecodeError(e)

    def read_little_endian64(self):
        """Interprets the next 8 bytes of the stream as a little-endian
        encoded, unsiged 64-bit integer, and returns that integer.
        """
        try:
            i = struct.unpack(wire_format.FORMAT_UINT64_LITTLE_ENDIAN,
                              self._input.read(8))
            self._pos += 8
            return i[0]  # unpack() result is a 1-element tuple.
        except struct.error as e:
            raise errors.DecodeError(e)

    def read_varint32(self):
        """Reads a varint from the stream, interprets this varint
        as a signed, 32-bit integer, and returns the integer.
        """
        i = self.read_varint64()
        if not wire_format.INT32_MIN <= i <= wire_format.INT32_MAX:
            raise errors.DecodeError('Value out of range for int32: %d' % i)
        return int(i)

    def read_var_uint32(self):
        """Reads a varint from the stream, interprets this varint
        as an unsigned, 32-bit integer, and returns the integer.
        """
        i = self.read_var_uint64()
        if i > wire_format.UINT32_MAX:
            raise errors.DecodeError('Value out of range for uint32: %d' % i)
        return i

    def read_varint64(self):
        """Reads a varint from the stream, interprets this varint
        as a signed, 64-bit integer, and returns the integer.
        """
        i = self.read_var_uint64()
        if i > wire_format.INT64_MAX:
            i -= (1 << 64)
        return i

    def read_var_uint64(self):
        """Reads a varint from the stream, interprets this varint
        as an unsigned, 64-bit integer, and returns the integer.
        """
        i = self._read_varint_helper()
        if not 0 <= i <= wire_format.UINT64_MAX:
            raise errors.DecodeError('Value out of range for uint64: %d' % i)
        return i

    def _read_varint_helper(self):
        """Helper for the various varint-reading methods above.
        Reads an unsigned, varint-encoded integer from the stream and
        returns this integer.

        Does no bounds checking except to ensure that we read at most as many bytes
        as could possibly be present in a varint-encoded 64-bit number.
        """
        result = 0
        shift = 0
        while 1:
            if shift >= 64:
                raise errors.DecodeError('Too many bytes when decoding varint.')
            try:
                b = ord(self._input.read(1))
            except IndexError:
                raise errors.DecodeError('Truncated varint.')
            self._pos += 1
            result |= ((b & 0x7f) << shift)
            shift += 7
            if not (b & 0x80):
                return result



# -*- coding: utf-8 -*-
import datetime

try:
    import zoneinfo
except ImportError:
    zoneinfo = None
try:
    import pytz
except ImportError:
    pytz = None


zone_not_found_errors = ()
if zoneinfo is not None:
    zone_not_found_errors += (zoneinfo.ZoneInfoNotFoundError,)
if pytz is not None:
    zone_not_found_errors += (pytz.UnknownTimeZoneError,)
if not zone_not_found_errors:
    raise ImportError("pytz not installed")


def get_system_offset():
    """Get system's timezone offset using built-in library time.

    For the Timezone constants (altzone, daylight, timezone, and tzname), the
    value is determined by the timezone rules in effect at module load time or
    the last time tzset() is called and may be incorrect for times in the past.

    To keep compatibility with Windows, we're always importing time module here.
    """
    import time
    if time.daylight and time.localtime().tm_isdst > 0:
        return -time.altzone
    else:
        return -time.timezone


def get_tz_offset(tz):
    """Get timezone's offset using built-in function datetime.utcoffset()."""
    return int(datetime.datetime.now(tz).utcoffset().total_seconds())


def assert_tz_offset(tz):
    """Assert that system's timezone offset equals to the timezone offset found.

    If they don't match, we probably have a misconfiguration, for example, an
    incorrect timezone set in /etc/timezone file in systemd distributions."""
    tz_offset = get_tz_offset(tz)
    system_offset = get_system_offset()
    if tz_offset != system_offset:
        msg = ('Timezone offset does not match system offset: {0} != {1}. '
               'Please, check your config files.').format(
                   tz_offset, system_offset
               )
        raise ValueError(msg)


def get_tz(zone):
    return zoneinfo.ZoneInfo(zone) if zoneinfo else pytz.timezone(zone)


def raise_zone_not_found(msg):
    if zoneinfo:
        raise zoneinfo.ZoneInfoNotFoundError(msg)
    else:
        raise pytz.UnknownTimeZoneError(msg)



# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

from ...utils import to_str


def _load_bearer_token_environ():
    from .cupid_service import CupidServiceClient

    if "CUPID_SERVICE_SOCKET" not in os.environ:
        return
    bearer_token = CupidServiceClient().get_bearer_token()
    os.environ["ODPS_BEARER_TOKEN"] = to_str(bearer_token)
    os.environ["ODPS_ENDPOINT"] = os.environ["ODPS_RUNTIME_ENDPOINT"]


_load_bearer_token_environ()
